"""This module computes general constrained equilibria from
`webnucleo <https://webnucleo.readthedocs.io>`_ files."""

import sys
from dataclasses import dataclass
import scipy.optimize as op
import numpy as np
import wnnet.consts as wc
import wneq.base as wqb


@dataclass
class _Cluster:
    name: str
    constraint: float
    mu: float
    index: int
    nuclides: list


class Equil(wqb.Base):
    """A class for handling constrained equilibria."""

    def __init__(self, nuc):
        wqb.Base.__init__(self, nuc)

        self.mup_kt = 0

    def compute(self, t9, rho, ye=None, clusters=None):
        """Method to compute a nuclear equilibrium.

        Args:
            ``t9`` (:obj:`float`): The temperature (in 10 :sup:`9` Kelvin)
            at which to compute the equilibrium.

            ``rho`` (:obj:`float`): The mass density in grams per cc  at which
            to compute the equilibrium.

            ``ye`` (:obj:`float`, optional): The electron fraction at which to compute
            the equilibrium.  If not supplied, the routine computes the equilibrium
            without a fixed total neutron-to-proton ratio.

            ``clusters`` (:obj:`dict`, optional): A dictionary with the key for each
            entry giving the XPath describing the cluster and the value giving the
            abundance constraint for the cluster.

        Returns:
            A `wnutils <https://wnutils.readthedocs.io>`_ zone data dictionary
            with the results of the calculation.

        """

        self.ye = ye
        self._update_fac(t9, rho)

        self.clusters.clear()
        if clusters:
            self._set_clusters(clusters)

        self._set_initial_guesses()

        x0 = self._get_initial_multi_vector()

        sol = op.root(self._compute_multi_root, x0)

        if (
            not sol.success
            or np.linalg.norm(self._compute_multi_root(sol.x)) > 1.0e-4
        ):
            res_bracket = self._bracket_root(
                self._compute_a_root, self.guess.mu["n"]
            )
            assert res_bracket, "Root not bracketed."
            res_root = op.root_scalar(
                self._compute_a_root, bracket=res_bracket
            )

            x0[0] = self.mup_kt
            x0[1] = res_root.root
            for value in self.clusters.values():
                x0[value.index] = value.mu

            sol = op.root(self._compute_multi_root, x0)

            assert sol.success, "Root polishing failed."

        self.mup_kt = sol.x[0]
        self.mun_kt = sol.x[1]
        for value in self.clusters.values():
            value.mu = sol.x[value.index]

        props = self._set_base_properties(t9, rho)

        props["mun_kT"] = self.mun_kt
        props["mup_kT"] = self.mup_kt
        props["mun"] = wc.ergs_to_MeV * (self.mun_kt * (wc.k_B * t9 * 1.0e9))
        props["mup"] = wc.ergs_to_MeV * (self.mup_kt * (wc.k_B * t9 * 1.0e9))

        for value in self.clusters.values():
            props[("cluster", value.name, "mu_kT")] = value.mu
            props[("cluster", value.name, "constraint")] = value.constraint

        y = self._compute_abundances(self.mup_kt, self.mun_kt)
        return self._make_equilibrium_zone(props, y)

    def _set_clusters(self, clusters):
        i = 2
        for key, value in clusters.items():
            self.clusters[key] = _Cluster(
                name=key,
                constraint=value,
                mu=0,
                index=i,
                nuclides=list(self.nuc.get_nuclides(nuc_xpath=key).keys()),
            )
            i += 1
        cluster_nuclide_set = set()
        for value in self.clusters.values():
            for nuc in value.nuclides:
                assert nuc not in cluster_nuclide_set
                cluster_nuclide_set.add(nuc)

    def _get_initial_multi_vector(self):
        n_var = 2
        if self.clusters:
            n_var += len(self.clusters)

        x0 = np.full(n_var, self.guess.x0)

        x0[0] = self.guess.mu["p"]
        x0[1] = self.guess.mu["n"]
        for key, value in self.clusters.items():
            x0[value.index] = self.guess.mu[key]

        return x0

    def _check_fac(self, _x):
        x_max = 300
        if isinstance(_x, float):
            return min(_x, x_max)

        for i, v_x in enumerate(_x):
            _x[i] = min(v_x, x_max)
        return _x

    def _compute_abundances(self, mup_kt, mun_kt):
        exp_fac = {}
        for key, value in self.nuc.get_nuclides().items():
            exp_fac[key] = (
                value["z"] * mup_kt
                + (value["a"] - value["z"]) * mun_kt
                + self.fac[key]
            )

        for value in self.clusters.values():
            for nuc in value.nuclides:
                exp_fac[nuc] += value.mu

        result = {}
        for nuc in self.nuc.get_nuclides():
            result[nuc] = np.exp(self._check_fac(exp_fac[nuc]))

        return result

    def _compute_multi_root(self, x):
        result = np.zeros(len(x))
        for key, value in self.clusters.items():
            value.mu = x[value.index]
            result[value.index] = value.constraint

        y = self._compute_abundances(x[0], x[1])

        result[0] = self.ye
        result[1] = 1

        for key, value in self.nuc.get_nuclides().items():
            result[0] -= value["z"] * y[key]
            result[1] -= value["a"] * y[key]

        for value in self.clusters.values():
            for nuc in value.nuclides:
                result[value.index] -= y[nuc]

        return result

    def _compute_a_root(self, x):

        if self.ye:
            res_bracket = self._bracket_root(
                self._compute_z_root, self.guess.mu["p"], args=(x,)
            )
            assert res_bracket, "Root not bracketed."
            res_root = op.root_scalar(
                self._compute_z_root, bracket=res_bracket, args=(x,)
            )
            self.mup_kt = res_root.root
        else:
            print("Feature not yet implemented.")
            sys.exit()

        y = self._compute_abundances(self.mup_kt, x)

        result = 1.0
        for key, value in self.nuc.get_nuclides().items():
            result -= value["a"] * y[key]

        return result

    def _compute_z_root(self, x, mun_kt):

        for key, value in self.clusters.items():
            res_bracket = self._bracket_root(
                self._compute_cluster_root,
                self.guess.mu[key],
                args=(x, mun_kt, key),
            )
            assert res_bracket, "Root not bracketed."
            res_root = op.root_scalar(
                self._compute_cluster_root,
                bracket=res_bracket,
                args=(x, mun_kt, key),
            )
            value.mu = res_root.root

        y = self._compute_abundances(x, mun_kt)

        result = self.ye
        for key, value in self.nuc.get_nuclides().items():
            result -= value["z"] * y[key]

        return result

    def _compute_cluster_root(self, x, mup_kt, mun_kt, cluster):

        self.clusters[cluster].mu = x
        y = self._compute_abundances(mup_kt, mun_kt)

        result = self.clusters[cluster].constraint
        for nuc in self.clusters[cluster].nuclides:
            result -= y[nuc]

        return result

    def compute_from_zone(self, zone, compute_ye=True, clusters=None):
        """Method to compute an equilibrium from input zone data.  The\
        resulting equilibrium is that to which the system would relax given\
        sufficient time.

        Args:
            ``zone``: A `wnutils <https://wnutils.readthedocs.io>`_ zone
            data dictionary with the physical conditions and abundances
            from which to compute the equilibrium.

            ``compute_ye`` (:obj:`bool`, optional): A boolean to determine whether to
            compute the electron fraction in the zone and use it for the equilibrium calculation.

            ``clusters`` (:obj:`list`, optional): A list of XPath strings describing the desired
            clusters for the equilibrium.

        Returns:
            A `wnutils <https://wnutils.readthedocs.io>`_ zone data
            dictionary with the results of the calculation.

        """

        t9 = float(zone["properties"]["t9"])
        rho = float(zone["properties"]["rho"])

        x_m = zone["mass fractions"]

        _y = {}

        ye = None
        if compute_ye:
            ye = 0

        for key, value in x_m.items():
            _y[key[0]] = value / key[2]
            if compute_ye:
                ye += key[1] * _y[key[0]]

        eq_clusters = None

        if clusters:
            eq_clusters = {}

            for cluster in clusters:
                y_c = 0
                for nuc in self.nuc.get_nuclides(nuc_xpath=cluster):
                    if nuc in _y:
                        y_c += _y[nuc]
                eq_clusters[cluster] = y_c

        return self.compute(t9, rho, ye=ye, clusters=eq_clusters)

    def compute_low_temperature_nse(self, ye=None):
        """Method to compute a nuclear statistical equilibrium at low\
        temperature in a one- or two-species approximation.

        Args:
            ``ye`` (:obj:`float`, optional): The electron fraction at which\
            to compute the equilibrium.  If not supplied, the routine computes\
            the equilibrium without a fixed total neutron-to-proton ratio,\
            in which case, the equilibrium is computed in a one-species\
            approximation.

        Returns:
            A `wnutils <https://wnutils.readthedocs.io>`_ zone data dictionary\
            with the results of the calculation.

        """
        if ye is not None:
            ye_t = ye
            min_pair = self._find_min_pair(ye)
            y = self._compute_pair_abundances(min_pair, ye_t)
        else:
            species = self._find_min_species()
            y = {species: 1.0 / self.nuc.get_nuclides()[species]["a"]}
            ye_t = self.nuc.get_nuclides()[species]["z"] * y[species]
        props = {
            "note": "computed in 1- or 2-species, low-temperature approximation",
            "Ye": ye_t,
        }
        return self._make_equilibrium_zone(props, y)

    def _compute_pair_ye_min(self, ye):
        if ye < 0 or ye > 1:
            return 1.0e6
        min_pair = self._find_min_pair(ye)
        y = self._compute_pair_abundances(min_pair, ye)
        return self._compute_mass_per_nucleon_for_pair(y)

    def _find_min_species(self):
        dm_min = np.inf
        result = ""
        for key, value in self.nuc.get_nuclides().items():
            dm = self.nuc.compute_atomic_mass(key) / value["a"]
            if dm < dm_min:
                dm_min = dm
                result = key
        return result

    def _find_min_pair(self, ye):
        nuc_list = list(self.nuc.get_nuclides().keys())
        dm_min = np.inf
        min_pair = (nuc_list[0], nuc_list[0])
        for i, first_nuc in enumerate(nuc_list):
            for j in range(i + 1, len(nuc_list)):
                y = self._compute_pair_abundances((first_nuc, nuc_list[j]), ye)
                dm = self._compute_mass_per_nucleon_for_pair(y)
                if dm < dm_min:
                    dm_min = dm
                    min_pair = (first_nuc, nuc_list[j])
        return min_pair

    def _compute_pair_abundances(self, pair, ye):
        nucs = self.nuc.get_nuclides()
        z_0 = nucs[pair[0]]["z"]
        a_0 = nucs[pair[0]]["a"]
        z_1 = nucs[pair[1]]["z"]
        a_1 = nucs[pair[1]]["a"]

        ye_0 = float(z_0) / float(a_0)
        ye_1 = float(z_1) / float(a_1)

        result = {}

        if ye_0 != ye_1:
            denom = ye_1 - ye_0
            result[pair[0]] = (1.0 / a_0) * (ye_1 - ye) / denom
            result[pair[1]] = (1.0 / a_1) * (ye - ye_0) / denom
            return result
        if ye_0 == ye:
            result[pair[0]] = 1.0 / a_0
            result[pair[1]] = 0.0
            return result
        return {pair[0]: -1.0, pair[1]: 1.0}

    def _compute_mass_per_nucleon_for_pair(self, y):
        for value in y.values():
            if value < 0:
                return np.inf

        result = 0
        for key, value in y.items():
            result += value * self.nuc.compute_atomic_mass(key)

        return result



"""This module computes (n,g)-(g,n) equilibrium from
`webnucleo <https://webnucleo.readthedocs.io>`_ files."""

import scipy.optimize as op
import numpy as np
import wnnet as wn
import wneq.base as wqb


class Ng(wqb.Base):
    """A class for handling (n,g)-(g,n) equilibria."""

    def _compute_ng(self, t9, rho, mun_kt, y_z):
        y_zt = {}
        y_l = {}
        ylm = {}
        mass_frac = {}
        props = self._set_base_properties(t9, rho)

        for key in y_z.keys():
            ylm[key] = float("-inf")
            y_zt[key] = 0

        for key, value in self.nuc.get_nuclides().items():
            if value["z"] in y_z:
                y_t = self.fac[key] + value["a"] * mun_kt
                if y_t > ylm[value["z"]]:
                    ylm[value["z"]] = y_t
                y_l[(key, value["z"], value["a"])] = y_t

        for key in y_l:
            y_l[key] -= ylm[key[1]]
            y_zt[key[1]] += np.exp(y_l[key])

        for key, value in y_z.items():
            props[("muz_kt", str(key))] = str(np.log(value / y_zt[key]))

        for key, value in y_l.items():
            s_z = str(key[1])
            mass_frac[key] = (
                np.exp(value + float(props[("muz_kt", s_z)])) * key[2]
            )

        for key in y_z.keys():
            muz_kt = float(props[("muz_kt", str(key))])
            props[("muz_kt", str(key))] = str(muz_kt - ylm[key])

        mass_frac[("n", 0, 1)] = np.exp(self.fac["n"] + mun_kt)

        props["mun_kt"] = str(mun_kt)

        props["mun"] = str(
            wn.consts.ergs_to_MeV * (mun_kt * (wn.consts.k_B * t9 * 1.0e9))
        )

        return {"properties": props, "mass fractions": mass_frac}

    def compute(self, t9, rho, mun, y_z):
        """Method to compute an (n,g)-(g,n) equilibrium.

        Args:
            ``t9`` (:obj:`float`): The temperature (in 10 :sup:`9` Kelvin)
            at which to compute the equilibrium.

            ``rho`` (:obj:`float`): The mass density in grams per cc  at which
            to compute the equilibrium.

            ``mun`` (:obj:`float`): The neutron chemical potential (in MeV)
            at which to compute the equilibrium..

            ``y_z`` (:obj:`dict`): A dictionary with the elemental abundances
            for the calculation.  The keys of the dictionary are :obj:`int`
            giving the atomic number while the value is the abundance per
            nucleon for that atomic number.  On successful return,
            the equilibrium abundances will have the same elemental abundances
            as those given in *y_z*.

        Returns:
            A `wnutils <https://wnutils.readthedocs.io>`_ zone data dictionary
            with the results of the calculation.

        """

        self._update_fac(t9, rho)

        mun_kt = mun * wn.consts.MeV_to_ergs / (wn.consts.k_B * t9 * 1.0e9)

        return self._compute_ng(t9, rho, mun_kt, y_z)

    def _root_func(self, x_var, t9, rho, y_z):

        result = 1

        n_g = self._compute_ng(t9, rho, x_var, y_z)

        x_m = n_g["mass fractions"]

        for x_t in x_m.values():
            result -= x_t

        return result

    def compute_with_root(self, t9, rho, y_z):
        """Method to compute an (n,g)-(g,n) equilibrium.  The resulting
        equilibrium is that the system would relax to in the absence of
        charge-changing reactions and given sufficient time.  The return
        result contains the neutron abundance and chemical potential for the
        appropriate equilibrium.

        Args:
            ``t9`` (:obj:`float`): The temperature (in 10 :sup:`9` Kelvin)
            at which to compute the equilibrium.

            ``rho`` (:obj:`float`): The mass density in grams per cc
            at which to compute the equilibrium.

            ``y_z`` (:obj:`dict`): A dictionary with the elemental abundances
            for the calculation.  The keys of the dictionary are :obj:`int`
            giving the atomic number while the value is the abundance per
            nucleon for that atomic number.  On successful return,
            the equilibrium abundances will have the save elemental abundances
            as those given in *y_z*.

        Returns:
            A `wnutils <https://wnutils.readthedocs.io>`_ zone data dictionary
            with the results of the calculation.

        """

        self._update_fac(t9, rho)

        self._set_initial_guesses()

        res_bracket = self._bracket_root(
            self._root_func, self.guess.mu["n"], args=(t9, rho, y_z)
        )
        res_root = op.root_scalar(
            self._root_func, bracket=res_bracket, args=(t9, rho, y_z)
        )

        result = self._compute_ng(t9, rho, res_root.root, y_z)

        return result

    def compute_with_root_from_zone(self, zone):
        """Method to compute an (n,g)-(g,n) equilibrium.  The resulting
        equilibrium is that the system would relax to in the absence of
        charge-changing reactions and given sufficient time.  The return
        result contains the neutron abundance and chemical potential for the
        appropriate equilibrium.

        Args:
            ``zone``: A `wnutils <https://wnutils.readthedocs.io>`_ zone
            data dictionary with the physical conditions and abundances
            from which to compute the equilibrium.

        Returns:
            A `wnutils <https://wnutils.readthedocs.io>`_ zone data
            dictionary with the results of the calculation.

        """

        t9 = float(zone["properties"]["t9"])
        rho = float(zone["properties"]["rho"])

        self._update_fac(t9, rho)

        x_m = zone["mass fractions"]

        y_z = {}

        for key, value in x_m.items():
            if key[1] != 0:
                if key[1] in y_z:
                    y_z[key[1]] += value / key[2]
                else:
                    y_z[key[1]] = value / key[2]

        self._set_initial_guesses()

        res_bracket = self._bracket_root(
            self._root_func, self.guess.mu["n"], args=(t9, rho, y_z)
        )
        res_root = op.root_scalar(
            self._root_func, bracket=res_bracket, args=(t9, rho, y_z)
        )

        result = self._compute_ng(t9, rho, res_root.root, y_z)

        return result



"""This module contains base elements for the equilibrium classes."""

from dataclasses import dataclass


@dataclass
class _Guess:
    x0: float
    user: {}
    mu: {}


class Base:
    """A class for handling data for the equilibrium codes.

    Args:
        ``nuc``: A wnnet \
        `nuclear data <https://wnnet.readthedocs.io/en/latest/wnnet.html#module-wnnet.nuc>`_\
        object.


    """

    def __init__(self, nuc):
        self.nuc = nuc
        self.fac = {}
        self.mun_kt = 0
        self.ye = None
        self.guess = _Guess(-10.0, {}, {})
        self.clusters = {}

    def get_nuclides(self, nuc_xpath=""):
        """Method to return a collection of nuclides.

        Args:
            ``nuc_xpath`` (:obj:`str`, optional): An XPath expression to
            select the nuclides.  Default is all species.

        Returns:
            A :obj:`dict` containing\
            `wnutils <https://wnutils.readthedocs.io>`_ nuclides.

        """

        return self.nuc.get_nuclides(nuc_xpath=nuc_xpath)

    def _update_fac(self, t9, rho):

        for nuc in self.nuc.get_nuclides():
            self.fac[nuc] = self.nuc.compute_nse_factor(nuc, t9, rho)

    def _bracket_root(self, f, x0, args=()):
        factor = 1.6
        max_iter = 1000
        x1 = x0
        x2 = x1 + 1
        f1 = f(x1, *args)
        f2 = f(x2, *args)
        for _ in range(max_iter):
            if f1 * f2 < 0:
                return (x1, x2)
            if abs(f1) < abs(f2):
                x1 += factor * (x1 - x2)
                f1 = f(x1, *args)
            else:
                x2 += factor * (x2 - x1)
                f2 = f(x2, *args)
        return None

    def update_initial_guesses(self, guesses):
        """Method to update initial guesses for chemical potentials divided by kT.

        Args:
            ``guesses`` (:obj:`dict`, optional): A dictionary of the guesses.  The allowed keys\
            of the dictionary are \"n\" (for the neutrons), \"p\" (for the protons), or\
            XPath expressions defining the valid clusters for the equilibrium.\
            The value for each key should be a float giving the initial guess for the\
            chemical potential divided by kT for the species or cluster.

        Returns:
            On successful return, the initial guesses for the species or clusters defined by the
            input keys are updated to the corresponding values.  These guesses will then be
            applied in the next calculation of the equilibrium.

        """

        self.guess.user.clear()
        for key, value in guesses.items():
            self.guess.user[key] = float(value)

    def _set_initial_guesses(self):
        self.guess.mu.clear()
        self.guess.mu["n"] = self.guess.x0
        self.guess.mu["p"] = self.guess.x0
        for cluster in self.clusters:
            self.guess.mu[cluster] = self.guess.x0

        for key, value in self.guess.user.items():
            if key in self.guess.mu:
                self.guess.mu[key] = value

    def _set_base_properties(self, t9, rho):
        result = {}
        result["t9"] = t9
        result["rho"] = rho
        if self.ye:
            result["ye"] = self.ye
        return result

    def _make_equilibrium_zone(self, props, y):
        mass_fracs = {}
        nucs = self.nuc.get_nuclides()
        for key, value in y.items():
            if value > 0:
                nuc = nucs[key]
                mass_fracs[(key, nuc["z"], nuc["a"])] = nuc["a"] * value

        return {"properties": props, "mass fractions": mass_fracs}
